from pydantic import BaseModel, HttpUrl

from .client import Empty, SDKClient, SDKResponse


class SendSmsRequest(BaseModel):
    source_address: str = "mobilmed"
    destination_address: str
    message: str


class SmsService:
    def __init__(self, client: SDKClient, url: HttpUrl, token: str):
        self._client = client
        self._url = url
        self._token = token

    def send_sms(self, query: SendSmsRequest, timeout=3) -> SDKResponse[Empty]:
        return self._client.post(
            self._url + "send/",
            Empty,
            json=query.dict(),
            timeout=timeout,
            headers={"Authorization": f"TokenService {self._token}"},
        )



from .barcode_recognizer import BarcodeRecognizeService, BarcodeRequest, BarcodeResponse
from .base import Gender
from .client import SDKClient
from .elk import ElkService, SearchClientRequest
from .face_detector import (
    FaceDetectorDetectRequest,
    FaceDetectorService,
    FindSimilarFaceRequest,
    FindSimilarFaceResponse,
)
from .mis.crm import (
    GetChangedOrgRequest,
    GetChangedOrgResponse,
    GetOrgInfoRequest,
    GetOrgInfoResponse,
    KonturApiService,
    SubscribeOrgRequestData,
    SubscribeOrgRequestParams,
    SubscribeOrgResponse,
)
from .mis.lab import (
    ConstraintRequiredField,
    ConstraintsRequest,
    ConstraintsResponse,
    ConstraintType,
    LabConstraintService,
)
from .mis.standalone import (
    ExportRequest,
    ExportResponse,
    GetLatestDumpRequest,
    GetLatestDumpResponse,
    ImportRequest,
    ImportResponse,
    LocalChangeResponse,
    MisStandAloneService,
)
from .mis.widgets import ManifestResponse, WidgetsService
from .mobile_backend import MobileBackendService
from .notemaster import NoteLogRequest, NoteMasterService, NoteStatusRequest
from .notification import (
    EmailConfigRequest,
    NotificationConfigRequest,
    NotificationRequest,
    NotificationService,
    TelegramConfigRequest,
)
from .pre_record import (
    FreeDatesRequest,
    FreeTimeRequest,
    GetPreRecordPlacesRequest,
    PreRecordConfirmPayment,
    PreRecordGetRequest,
    PreRecordRequest,
    PreRecordService,
)
from .sms import SendSmsRequest, SmsService



from typing import List, Optional
from urllib.parse import urljoin

from pydantic import BaseModel, Field, HttpUrl

from .client import Empty, SDKClient


class AdditionalFields(BaseModel):
    mc_code: str
    pre_record_number: Optional[str]


class Product(BaseModel):
    id: str
    name: str
    quantity: int = 1
    price: int


class Order(BaseModel):
    id: str
    name: str
    date_create: str
    status: str = "1"  # взял из примера, нет описания что такое статус
    price: str
    client_id: str
    roistat: Optional[str]
    additional_fields: AdditionalFields = Field(alias="fields")
    products: List[Product]


class Orders(BaseModel):
    __root__: List[Order]


class Auth(BaseModel):
    project: str
    key: str


class RoistatService:
    def __init__(self, client: SDKClient, url: HttpUrl, auth: Auth):
        assert url.endswith("/")
        self._client = client
        self._url = url
        self._auth = auth

    def add_orders(self, query: Orders, timeout: int = 3):
        return self._client.post(
            urljoin(str(self._url), "api/v1/project/add-orders"),
            Empty,
            params=self._auth.dict(),
            data=query.json(by_alias=True),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )



import datetime
import json

from decimal import Decimal
from enum import Enum
from typing import Dict, List, Optional, Union
from urllib.parse import urljoin

from pydantic import BaseModel, HttpUrl, IPvAnyAddress, Json, root_validator

from . import Gender
from .client import SDKClient, SDKResponse


class ServiceForOrder(int, Enum):
    LAB = 5
    LMK = 4
    PROF = 3
    CERT = 1
    OUT = 6


class PaySystem(str, Enum):
    OTHER = ""  # когда наличка, не передается система оплаты
    TINKOFF_PAYMENT = "tinkoff"
    YOOKASSA_PAYMENT = "yookassa"
    # деньги платятся яндексу. Мис считает заказ полностью оплаченным
    YANDEX_PAYMENT = "yandex"


class PassportType(str, Enum):
    PASSPORT = "Паспорт гражданина РФ"
    ABROAD = "Загранпаспорт гражданина РФ"
    OTHER = "Паспорт иностранного гражданина"


class CyclePhase(str, Enum):
    FOLLICULAR = "follicular"
    OVULATORY = "ovulatory"
    LUTEAL = "luteal"
    MENOPAUSE = "menopause"
    POSTMENOPAUSE = "post_menopausal"
    PREMENOPAUSE = "premenopausal"
    NOPREGNANCY = "no pregnancy"


class CovidFields(BaseModel):
    gender: Optional[Gender]
    passport_type: Optional[PassportType]
    passport_date: Optional[datetime.date]
    passport_code: Optional[str]
    passport_series: Optional[str]
    passport_number: Optional[str]
    oms: Optional[str]
    snils: Optional[str]
    is_moscow: Optional[bool]
    registration_region: Optional[str] = ""
    registration_city: Optional[str]
    registration_street: Optional[str]
    registration_building: Optional[str] = ""
    registration_quarter: Optional[str] = ""
    fact_country: Optional[str]
    fact_republic: Optional[str]
    fact_region: Optional[str]
    fact_city: Optional[str]
    fact_street: Optional[str]
    fact_building: Optional[str]
    fact_quarter: Optional[str]
    vaccine_covid_first: Optional[datetime.date]
    vaccine_covid_first_name: Optional[str]
    vaccine_covid_second: Optional[datetime.date]
    vaccine_covid_second_name: Optional[str]

    class Config:
        use_enum_values = True


class OutData(BaseModel):
    address: str
    lat: Optional[str]
    long: Optional[str]


class LabExtraFields(BaseModel):
    gender: Optional[Gender]
    nationality: Optional[str]
    passport_type: Optional[PassportType]
    passport_date: Optional[str]
    passport_code: Optional[str]
    passport_series: Optional[str]
    passport_number: Optional[str]
    registration_city: Optional[str]
    registration_street: Optional[str]
    registration_building: Optional[str]
    registration_quarter: Optional[str]
    diuresis: Optional[str]
    weight: Optional[int]
    height: Optional[int]
    pregnancy_week: Optional[int]
    cycle_day: Optional[int]
    cycle_phase: Optional[CyclePhase]
    work_place: Optional[str]
    work_place_phone: Optional[str]
    work_place_address: Optional[str]
    trip: Optional[str]
    out: Optional[OutData]

    class Config:
        use_enum_values = True


class Source(str, Enum):
    site = "site"
    call_center = "call_center"
    delivery = "delivery"
    mobile = "mobile"


class PayType(str, Enum):
    cashless_prepayment = "cashless_prepayment"
    med_center = "med_center"
    delivery = "delivery"


class FreeDatesRequest(BaseModel):
    services: List[int]
    lab_codes: Optional[List[str]] = None
    place: int
    dt_from: datetime.datetime


class FreeTimeRequest(BaseModel):
    services: List[int]
    lab_codes: Optional[List[str]] = None
    place: int
    date: datetime.date


class GetPreRecordPlacesRequest(BaseModel):
    lab_codes: Optional[List[str]] = None


class Service(BaseModel):
    name: str
    slug: str


class WorkSchedule(BaseModel):
    week_day: int
    time_from: datetime.time
    time_to: datetime.time
    week_day_name: str


class PlacesResponse(BaseModel):
    id: int
    name: str
    med_center_id: int
    metro: str
    extra_metro: List[str]
    subnet: IPvAnyAddress
    med_center_extra_info: Union[Dict, Json]
    lat: Decimal
    lng: Decimal


class RecordTime(BaseModel):
    dt_from: datetime.datetime
    dt_to: datetime.datetime


class FreeTimeResponse(BaseModel):
    service: Service
    records: List[RecordTime]


class PreRecordRequest(BaseModel):
    service: int
    place: int
    date: datetime.date
    time: datetime.time
    birth: Optional[datetime.date]
    last_name: str
    first_name: str
    middle_name: str = ""
    client_uuid: str = ""
    operator_login: Optional[str]
    email: str
    phone: str
    analysis: List[str] = []
    source: Source = Source.site
    lab_extra_fields: Optional[LabExtraFields]
    covid_fields: Optional[CovidFields]
    pay_type: PayType = PayType.med_center
    called_from_number: Optional[str]
    operator_comment: Optional[str]
    ga_client_id: Optional[str]
    ym_client_id: Optional[str]
    admitad_code: Optional[str]
    pay_system: Optional[PaySystem]
    roistat: Optional[str]
    # игнорирую, потому что пока ненужно. Но в апи передают
    roistat_first: Optional[str]
    with_discount: bool = True
    gift_code: str = ""
    initial_cost: Optional[int]

    class Config:
        use_enum_values = True

    def convert_to_request(self):
        """Преобразовывает данные в формат, который принимает предварительная запись."""
        data = {
            "lname": self.last_name,
            "fname": self.first_name,
            "mname": self.middle_name,
            "operator_login": self.operator_login,
            "birth": self.birth.strftime("%Y-%m-%d") if self.birth else None,
            "phone": self.phone,
            "d_from": datetime.datetime.combine(self.date, self.time).strftime(
                "%Y-%m-%d %H:%M"
            ),
            "d_to": (
                datetime.datetime.combine(self.date, self.time)
                + datetime.timedelta(minutes=10)
            ).strftime("%Y-%m-%d %H:%M"),
            "place": self.place,
            "email": self.email,
            "services": [{"id": self.service}],
            "lab_services": [
                {"price_code": price_code} for price_code in self.analysis
            ],
            "source": self.source,
            "pay_type": self.pay_type,
            "with_discount": self.with_discount,
            "gift_code": self.gift_code,
        }
        if self.initial_cost:
            data["initial_cost"] = self.initial_cost

        if self.lab_extra_fields:
            data["extra_fields"] = self.lab_extra_fields.dict(exclude_none=True)
        if self.covid_fields:
            # преобразую, в json и обратно, потому что иначе datetime остается в
            # словаре и http json дампер не переваривает
            data["covid_fields"] = json.loads(self.covid_fields.json(exclude_none=True))
        if self.called_from_number:
            data["called_from_number"] = self.called_from_number
        if self.operator_comment:
            data["operator_comment"] = self.operator_comment
        if self.ym_client_id:
            data["ym_client_id"] = self.ym_client_id
        if self.ga_client_id:
            data["ga_client_id"] = self.ga_client_id
        if self.roistat:
            data["roistat"] = self.roistat
        if self.admitad_code:
            data["admitad_code"] = self.admitad_code
        if self.pay_system:
            data["pay_system"] = self.pay_system
        if self.client_uuid:
            data["client_uuid"] = self.client_uuid
        return data


class DeliveryPreRecordRequest(BaseModel):
    birth: datetime.date
    last_name: str
    first_name: str
    middle_name: str = ""
    email: str
    phone: str
    covid_fields: CovidFields

    class Config:
        use_enum_values = True


class PreRecordTickRequest(BaseModel):
    # забрал со старого апи, это результат json.dumps() примерный набор полей
    # phone/cost/lab_services(cost/price_code
    external_data: bytes
    lmk_order_uuid: Optional[str]
    lab_order_uuid: Optional[str]
    prof_order_uuid: Optional[str]
    certificate_order_uuid: Optional[str]
    client_uuid: Optional[str]
    lname: Optional[str]
    mname: Optional[str]
    fname: Optional[str]
    birth: Optional[datetime.date]


class ClientPreRecordsRequest(BaseModel):
    client_uuid: Optional[str]
    phone: Optional[str]
    d_to: Optional[datetime.date]
    lname: Optional[str]
    services: Optional[List[ServiceForOrder]]
    limit: int = 10
    offset: int = 0

    class Config:
        use_enum_values = True

    @root_validator(pre=True)
    def one_of(cls, values):
        client_uuid = values.get("client_uuid")
        phone = values.get("phone")
        if not client_uuid and not phone:
            raise ValueError("phone or client_uuid is required")
        return values


class PreRecordGetRequest(BaseModel):
    number: int
    client_uuid: Optional[str]
    phone: Optional[str]


class LabService(BaseModel):
    price_code: str


class OrderResults(BaseModel):
    lab_results_url: Optional[str]
    covid_results_url: Optional[str]
    flg_results_url: Optional[str]


class PreRecordResponse(BaseModel):
    number: int
    phone: str
    services: List[Service]
    lab_services: Optional[List[Optional[LabService]]]
    d_from: datetime.datetime
    d_to: datetime.datetime
    prepayment_is_paid: bool
    prepayment_cost: Optional[int]
    pay_type: PayType
    visit_cost: Optional[int]
    initial_cost: Optional[int]
    pay_system: Optional[PaySystem]
    pay_service_id: Optional[str]
    med_center_id: int
    lname: str
    fname: str
    mname: str
    birth: Optional[datetime.date]
    client_uuid: Optional[str]
    email: Optional[str]
    gift_code: Optional[str]
    place: int
    extra_fields: Optional[LabExtraFields]
    covid_fields: Optional[CovidFields]
    results: Optional[OrderResults]
    dc: datetime.datetime

    class Config:
        use_enum_values = True


class PaginatedPreRecordResponse(BaseModel):
    count: int
    next: Optional[str]
    previous: Optional[str]
    results: List[PreRecordResponse]


class PreRecordDeliveryResponse(BaseModel):
    number: int


class PreRecordConfirmPayment(BaseModel):
    number: int
    pay_service_id: str
    prepayment_cost: str
    pay_system: Optional[PaySystem]

    class Config:
        use_enum_values = True


class AnalyticTags(BaseModel):
    number: int
    roistat: str
    roistat_first: Optional[str]


class EmptyResponse(BaseModel):
    pass


class PreRecordService:
    def __init__(self, client: SDKClient, url: HttpUrl):
        assert url.endswith("/")
        self._client = client
        self._url = url

    def places(
        self, query: Optional[GetPreRecordPlacesRequest] = None, timeout=15
    ) -> SDKResponse[List[PlacesResponse]]:
        return self._client.get(
            urljoin(str(self._url), "rest/place/"),
            PlacesResponse,
            params={"lab_codes": query.lab_codes if query else None},
            timeout=timeout,
        )

    def free_dates(self, query: FreeDatesRequest, timeout=15) -> SDKResponse[list]:
        return self._client.get(
            urljoin(str(self._url), "rest/pre_record/free_dates/"),
            list,
            params={
                "services": query.services,
                "lab_codes": query.lab_codes,
                "place": query.place,
                "dt_from": query.dt_from.isoformat()[:19],
            },
            timeout=timeout,
        )

    def free_time(
        self, query: FreeTimeRequest, timeout=15
    ) -> SDKResponse[List[FreeTimeResponse]]:
        return self._client.get(
            urljoin(str(self._url), "rest/pre_record/free/"),
            FreeTimeResponse,
            params={
                "services": query.services,
                "lab_codes": query.lab_codes,
                "place": query.place,
                "date": query.date.isoformat(),
            },
            timeout=timeout,
        )

    def create_pre_record(
        self, query: PreRecordRequest, timeout=15
    ) -> SDKResponse[PreRecordResponse]:
        return self._client.post(
            urljoin(str(self._url), "rest/pre_record/"),
            PreRecordResponse,
            data=json.dumps(query.convert_to_request()),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def create_delivery_pre_record(
        self, query: DeliveryPreRecordRequest, timeout=15
    ) -> SDKResponse[PreRecordResponse]:
        """Создание предварительной записи по "доставке".
        Интеграция с анкетой на сайте mobil-med.org для ковид-пцр и яндекс доставкй
        """
        data = {
            "lname": query.last_name,
            "fname": query.first_name,
            "mname": query.middle_name,
            "birth": query.birth.strftime("%Y-%m-%d"),
            "phone": query.phone,
            "email": query.email,
            "covid_fields": json.loads(query.covid_fields.json(exclude_none=True)),
        }

        return self._client.post(
            f"{self._url}rest/pre_record_delivery/",
            PreRecordDeliveryResponse,
            data=json.dumps(data),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def update_pre_record(
        self, number, query: PreRecordRequest, timeout=15
    ) -> SDKResponse[PreRecordResponse]:
        return self._client.put(
            urljoin(str(self._url), f"rest/pre_record/{number}/"),
            PreRecordResponse,
            data=json.dumps(query.convert_to_request()),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def confirm_payment(
        self, query: PreRecordConfirmPayment, timeout=15
    ) -> SDKResponse[PreRecordResponse]:
        data = {
            "prepayment_cost": query.prepayment_cost,
            "prepayment_is_paid": True,
            "pay_service_id": query.pay_service_id,
        }
        if query.pay_system:
            data["pay_system"] = query.pay_system

        return self._client.patch(
            f"{self._url}rest/pre_record/{query.number}",
            PreRecordResponse,
            data=json.dumps(data),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def set_analytics(
        self, query: AnalyticTags, timeout=15
    ) -> SDKResponse[PreRecordResponse]:
        return self._client.patch(
            f"{self._url}rest/pre_record/{query.number}",
            PreRecordResponse,
            data=json.dumps({"roistat": query.roistat}),
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def get_pre_record(
        self, query: PreRecordGetRequest, timeout=10, strict: bool = False
    ) -> SDKResponse[PreRecordResponse]:
        url = f"{self._url}rest/pre_record/{query.number}"
        params = query.dict()
        del params["number"]
        if strict:
            url = f"{self._url}rest/pre_record_strict/{query.number}"
        return self._client.get(
            url,
            PreRecordResponse,
            params=params,
            timeout=timeout,
            headers={"Content-Type": "application/json"},
        )

    def get_client_pre_records(
        self,
        query: ClientPreRecordsRequest,
        with_pagination: bool = False,
        timeout: int = 10,
        strict: bool = False,
    ) -> SDKResponse[Union[List[PreRecordResponse], PaginatedPreRecordResponse]]:
        url = f"{self._url}rest/pre_record/"
        if strict:
            url = f"{self._url}rest/pre_record_strict/"
        resp: SDKResponse[PaginatedPreRecordResponse] = self._client.get(
            url,
            PaginatedPreRecordResponse,
            timeout=timeout,
            params=query.dict(),
            headers={"Content-Type": "application/json"},
        )
        if with_pagination:
            return resp
        else:
            # старый формат
            resp = SDKResponse[List[PreRecordResponse]](data=resp.data.results)
            return resp

    def set_tick(
        self, number: int, query: PreRecordTickRequest, timeout: int = 2
    ) -> SDKResponse[PreRecordResponse]:
        return self._client.patch(
            f"{self._url}rest/pre_record/{number}/set_tick/",
            PreRecordResponse,
            timeout=timeout,
            params=query.dict(exclude_unset=True),
            headers={"Content-Type": "application/json"},
        )



import datetime

from typing import Optional
from urllib.parse import urljoin

from pydantic import BaseModel, Field, HttpUrl

from .. import Gender
from ..client import SDKClient, SDKResponse


MC_URL_BY_ID = {
    1: "https://sm.mm/",
    2: "https://dm.mm/",
    3: "https://tm.mm/",
    4: "https://fm.mm/",
}


class GetClient(BaseModel):
    uuid: str


class Client(BaseModel):
    uuid: str
    phone: str
    last_name: str = Field(alias="lname")
    first_name: str = Field(alias="fname")
    middle_name: Optional[str] = Field(alias="mname", default="")
    birth: datetime.date
    email: Optional[str]
    sex: Gender

    class Config:
        use_enum_values = True


class ClientResultsRequest(BaseModel):
    client_uuid: str
    mc_id: int
    pre_record_number: int
    lab_number: int


class ClientResultsResponse(BaseModel):
    lab_results_url: Optional[str]
    covid_results_url: Optional[str]
    flg_results_url: Optional[str]


class ClientService:
    def __init__(self, client: SDKClient, url: HttpUrl):
        self._client = client
        self._url = url

    def get_client(self, query: GetClient, timeout=3) -> SDKResponse[Client]:
        return self._client.get(
            urljoin(str(self._url), f"mobil/rest/client/{query.uuid}/"),
            Client,
            timeout=timeout,
        )

    def get_results(self, query: ClientResultsRequest, timeout=3):
        params = {
            "lab_number": query.lab_number,
            "pre_record_number": query.pre_record_number,
        }
        return self._client.get(
            urljoin(
                MC_URL_BY_ID[query.mc_id], f"mobil/client/{query.client_uuid}/results/"
            ),
            ClientResultsResponse,
            params=params,
            timeout=timeout,
        )



import enum
from typing import Optional, Union

from pydantic import BaseModel

from .client import HttpUrl, SDKClient, SDKResponse


class EmailConfigRequest(BaseModel):
    name: str
    host_name: str
    host_port: str
    username: str
    password: str


class EmailConfigResponse(BaseModel):
    created: str


class TelegramConfigRequest(BaseModel):
    name: str
    bot_name: str
    token: str


class TelegramConfigResponse(BaseModel):
    created: str


class NotificationConfigRequest(BaseModel):
    service: str
    notification_type: str
    email_config: Optional[int] = None
    telegram_config: Optional[int] = None


class NotificationConfigResponse(BaseModel):
    created: str
    error: Optional[str] = None


class NotificationRequest(BaseModel):
    msg: str
    receiver: str
    subject: Optional[str] = None
    service: str
    notification_type: str
    spam_countdown: Optional[int] = None  # время в секундах


class NotificationResponse(BaseModel):
    status: str
    error: Optional[str] = None


class NotificationService:
    def __init__(self, client: SDKClient, url: HttpUrl):
        self._client = client
        self._url = url
        self._url_email_config = self._url + "api/email_config"
        self._url_telegram_config = self._url + "api/telegram_config"
        self._url_notification_config = self._url + "api/notification_config"
        self._url_notification = self._url + "api/notification"

    def create_email_config(
        self, query: EmailConfigRequest, timeout=3
    ) -> SDKResponse[EmailConfigResponse]:
        return self._client.post(
            self._url_email_config,
            response_class=EmailConfigResponse,
            data=query.json(),
            headers={"Content-Type": "application/json"},
            timeout=timeout,
        )

    def create_telegram_config(
        self, query: TelegramConfigRequest, timeout=3
    ) -> SDKResponse[TelegramConfigResponse]:
        return self._client.post(
            self._url_telegram_config,
            response_class=TelegramConfigResponse,
            data=query.json(),
            headers={"Content-Type": "application/json"},
            timeout=timeout,
        )

    def create_notification_config(
        self, query: NotificationConfigRequest, timeout=3
    ) -> SDKResponse[NotificationConfigResponse]:
        return self._client.post(
            self._url_notification_config,
            response_class=NotificationConfigResponse,
            data=query.json(),
            timeout=timeout,
        )

    def create_notification(
        self, query: NotificationRequest, timeout=3
    ) -> SDKResponse[NotificationResponse]:
        return self._client.post(
            self._url_notification,
            response_class=NotificationResponse,
            data=query.json(),
            timeout=timeout,
        )



import datetime

from typing import List, Optional

from pydantic import BaseModel, HttpUrl

from .client import SDKClient, SDKResponse


class GiftCodesRequest(BaseModel):
    code: str = None
    not_expired: bool = True
    is_promo: bool = True


class GiftCodeRequest(BaseModel):
    code: str
    not_expired: bool = True
    is_promo: bool = True


class GiftCodeResponse(BaseModel):
    code: str
    amount: int
    dc: datetime.datetime
    expired_date: datetime.datetime
    used_date: Optional[datetime.datetime]
    is_reusable: bool
    is_promo: bool
    min_cost: Optional[int]


class GiftService:
    def __init__(self, client: SDKClient, url: HttpUrl):
        self._client = client
        self._url = url

    def get_codes(
        self, query: GiftCodesRequest = GiftCodesRequest(), timeout=3
    ) -> SDKResponse[List[GiftCodeResponse]]:
        return self._client.get(
            self._url + "gifts/rest/",
            GiftCodeResponse,
            params=query.dict(exclude_none=True),
            timeout=timeout,
        )

    def get_code(
        self, query: GiftCodeRequest, timeout=3
    ) -> SDKResponse[GiftCodeResponse]:
        params = query.dict(exclude_none=True)
        code = params.pop("code")
        return self._client.get(
            self._url + f"gifts/rest/{code}/",
            GiftCodeResponse,
            params=params,
            timeout=timeout,
        )



"""
    Typed http client with error handling.

    # >>> import logging, pydantic
    # >>> logging.basicConfig(level=logging.DEBUG)
    # >>> client = SDKClient(connection_timeout=10, io_timeout=10, silent=True)
    # >>> class UserAgent(pydantic.BaseModel):
    # ...     user_agent: str = pydantic.Field(None, alias='user-agent')
    # >>> response = client.get('http://httpbin.org/user-agent', UserAgent)
    # >>> response
    # SDKResponse[Union[List[sdk.client.UserAgent], UserAgent]](data=UserAgent(user_agent='mobil-sdk/0.0.1'), error=None)
    #
    # >>> response = client.get('http://httpbin.org/post', UserAgent)
    # >>> response
    # SDKResponse[str](data=None, error=Error(code=405, message='METHOD NOT ALLOWED', detail='<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\\n<title>405 Method Not Allowed</title>\\n<h1>Method Not Allowed</h1>\\n<p>The method is not allowed for the requested URL.</p>\\n'))

"""
import inspect
import json
import logging

from http import HTTPStatus
from typing import (
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from httpx import Client, Response, Timeout
from pydantic import BaseModel, HttpUrl
from pydantic.generics import GenericModel

from sdk.base import BytesResponse


logger = logging.getLogger("sdk")


DataT = TypeVar("DataT")


class Error(BaseModel):
    code: int
    message: str
    detail: Optional[str] = None


class SDKResponse(GenericModel, Generic[DataT]):
    data: Optional[DataT]
    error: Optional[Error]


class Empty(BaseModel):
    pass


class SDKClient:
    def __init__(
        self,
        connection_timeout: Union[int, float] = None,  # seconds, required
        io_timeout: Union[int, float] = None,  # seconds, required
        client: Optional[Client] = None,
        headers: Union[Dict, Callable] = None,
        acceptable_status_codes: Sequence = (
            HTTPStatus.OK,
            HTTPStatus.CREATED,
            HTTPStatus.ACCEPTED,
        ),
        silent: bool = True,
        content_slice_size: int = 256,
    ):
        assert client is not None or io_timeout, "Please provide io_timeout explicitly"
        assert (
            client is not None or connection_timeout is not None
        ), "session or connection_timeout must be provided"

        self._connection_timeout = connection_timeout
        self._io_timeout = io_timeout
        self._headers = headers or {}
        self._acceptable_status_codes = acceptable_status_codes

        self._client = client or self.create_default_client(
            self._connection_timeout, self._io_timeout
        )
        self._silent = silent
        self._content_slice_size = content_slice_size

    def __del__(self):
        self.close()

    @staticmethod
    def create_default_client(
        connection_timeout: Union[float, int, None], io_timeout: Union[float, int, None]
    ) -> Client:
        timeout = Timeout(io_timeout, connect=connection_timeout)
        return Client(
            headers={"user-agent": "mobil-sdk/0.0.1"}, timeout=timeout, verify=False
        )

    def _update_request_parameters(self, params: dict) -> dict:
        kwargs = params.copy()
        if callable(self._headers):
            default_headers = self._headers()
        else:
            default_headers = self._headers
        if "headers" in kwargs:
            new_headers = default_headers.copy()

            new_headers.update(kwargs["headers"])
            kwargs["headers"] = new_headers
        else:
            kwargs["headers"] = default_headers.copy()

        kwargs.setdefault("timeout", self._io_timeout)

        return kwargs

    def request(
        self, method: str, url: HttpUrl, response_class: Type[BaseModel], **kwargs
    ) -> SDKResponse:

        try:
            params = self._update_request_parameters(kwargs)

            logger.info(
                "~~~> HTTP request method=%s, url=%s, params=%s",
                method,
                url,
                self._log_kwargs(params),
            )

            response = self._client.request(method, url, **params)

            logger.info(
                "<~~~ HTTP response status_code=%s, url=%s content=%s",
                response.status_code,
                response.url,
                response.content[: self._content_slice_size],
            )

        except Exception as ex:
            if not self._silent:
                raise

            logger.error(
                "<~~~ Request error: %s method: %s  url: %s  kwargs: %s",
                ex,
                method,
                url,
                self._log_kwargs(kwargs),
                exc_info=True,
            )
            return SDKResponse[str](
                error={
                    "code": ex.status_code  # type: ignore
                    if hasattr(ex, "status_code")
                    else HTTPStatus.INTERNAL_SERVER_ERROR,
                    "message": str(ex),
                }
            )

        if (
            self._acceptable_status_codes
            and response.status_code not in self._acceptable_status_codes
        ):
            if not self._silent:
                response.raise_for_status()
            logger.error(
                "<~~~ Unexpected status_code=%s received from %s url=%s kwargs: %s",
                response.status_code,
                method,
                url,
                self._log_kwargs(params),
                exc_info=True,
            )
            return SDKResponse[str](
                error={
                    "code": response.status_code,
                    "message": response.reason_phrase or "",
                    "detail": response.content
                    if hasattr(response, "content")
                    else None,
                }
            )

        return SDKResponse[Union[List[response_class], response_class]](
            data=self._to_sdk_response(response, response_class)
        )

    def _log_kwargs(self, params):
        log_params = params.copy()
        log_params.pop("headers", None)
        log_params.pop("files", None)
        return json.dumps(log_params, default=str)[: self._content_slice_size]

    @staticmethod
    def _to_sdk_response(
        response: Response, response_class: Type[BaseModel]
    ) -> Union[List[BaseModel], BaseModel]:
        if response_class == Empty:
            return None

        if inspect.isclass(response_class) and issubclass(
            response_class, BytesResponse
        ):
            return response_class(bytes=response.content)
        else:
            response_json = json.loads(response.content)
            if isinstance(response_json, list):
                if len(response_json) and isinstance(response_json[0], dict):
                    return [response_class(**x) for x in response_json]
                else:
                    return response_json
            else:
                return response_class(**response_json)

    def get(
        self, url, response_class: Union[Type[BaseModel], Type[List]], **kwargs
    ) -> SDKResponse:
        return self.request("GET", url, response_class, **kwargs)

    def post(self, url, response_class: Type[BaseModel], **kwargs) -> SDKResponse:
        return self.request("POST", url, response_class, **kwargs)

    def put(self, url, response_class: Type[BaseModel], **kwargs) -> SDKResponse:
        return self.request("PUT", url, response_class, **kwargs)

    def patch(self, url, response_class: Type[BaseModel], **kwargs) -> SDKResponse:
        return self.request("PATCH", url, response_class, **kwargs)

    def delete(self, url, response_class: Type[BaseModel], **kwargs) -> SDKResponse:
        return self.request("DELETE", url, response_class, **kwargs)

    def close(self):
        if hasattr(self, "_client"):
            self._client.close()


if __name__ == "__main__":
    from doctest import testmod

    testmod(verbose=False)
